@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2017  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2017-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -427,7 +427,7 @@ static void modem_destroy(void *data)
 		iwd_sim_auth_remove(modem->auth);
 	}
 
-	l_debug("removing modem %s\n", modem->path);
+	l_debug("removing modem %s", modem->path);
 
 	if (modem->apps_serial)
 		l_dbus_cancel(dbus, modem->apps_serial);
