@@ -253,7 +253,7 @@ static bool eap_mschapv2_send_response(struct eap_state *eap)
 							MSCHAPV2_CHAL_LEN);
 	memcpy(response->name, state->user, state->user_len);
 
-	eap_send_response(eap, EAP_TYPE_MSCHAPV2, output, sizeof(output));
+	eap_method_respond(eap, output, sizeof(output));
 
 	return true;
 }
@@ -346,7 +346,7 @@ static void eap_mschapv2_handle_success(struct eap_state *eap,
 	eap_method_success(eap);
 
 	buffer[5] = MSCHAPV2_OP_SUCCESS;
-	eap_send_response(eap, EAP_TYPE_MSCHAPV2, buffer, sizeof(buffer));
+	eap_method_respond(eap, buffer, sizeof(buffer));
 
 	/* The eapol set_key_material only needs msk, and that's all we got */
 	eap_set_key_material(eap, session_key, 32, NULL, 0, NULL, 0, NULL, 0);
